@@ -3,11 +3,13 @@
 #
 
 import re
-from Errors import error, InternalError
+from Errors import warning, error, InternalError
 import Options
 import Naming
-from PyrexTypes import c_int_type, \
-    py_object_type, c_char_array_type, \
+import PyrexTypes
+from PyrexTypes import \
+    py_object_type, py_type_type, \
+    c_int_type, c_char_array_type, \
     CEnumType, CStructOrUnionType, PyExtensionType
 from TypeSlots import \
     pyfunction_signature, pymethod_signature, \
@@ -18,60 +20,68 @@ identifier_pattern = re.compile(r"[A-Za-z_][A-Za-z0-9_]*$")
 class Entry:
     # A symbol table entry in a Scope or ModuleNamespace.
     #
-    # name             string     Python name of entity
-    # cname            string     C name of entity
-    # type             PyrexType  Type of entity
-    # doc              string     Doc string
-    # init             string     Initial value
-    # visibility       'private' or 'public' or 'extern'
-    # is_builtin       boolean    Is a Python builtin name
-    # is_cglobal       boolean    Is a C global variable
-    # is_pyglobal      boolean    Is a Python module-level variable
-    #                               or class attribute during
-    #                               class construction
-    # is_variable      boolean    Is a variable
-    # is_cfunction     boolean    Is a C function
-    # is_cmethod       boolean    Is a C method of an extension type
-    # is_type          boolean    Is a type definition
-    # is_const         boolean    Is a constant
-    # is_property      boolean    Is a property of an extension type:
-    # doc_cname        string or None  C const holding the docstring
-    # getter_cname     string          C func for getting property
-    # setter_cname     string          C func for setting or deleting property
-    # is_self_arg      boolean    Is the "self" arg of an exttype method
-    # is_readonly      boolean    Can't be assigned to
-    # func_cname       string     C func implementing Python func
-    # pos              position   Source position where declared
-    # namespace_cname  string     If is_pyglobal, the C variable
-    #                               holding its home namespace
-    # pymethdef_cname  string     PyMethodDef structure
-    # signature        Signature  Arg & return types for Python func
-    # init_to_none     boolean    True if initial value should be None
-    # as_variable      Entry      Alternative interpretation of extension
-    #                               type name as a variable
-    # xdecref_cleanup  boolean    Use Py_XDECREF for error cleanup
-    # in_cinclude      boolean    Suppress C declaration code
-    # enum_values      [Entry]    For enum types, list of values
-    # qualified_name   string     "modname.funcname" or "modname.classname"
-    #                               or "modname.classname.funcname"
-    # is_declared_generic  boolean  Is declared as PyObject * even though its
-    #                                 type is an extension type
-    # as_module        None       Module scope, if a cimported module
-    # is_inherited     boolean    Is an inherited attribute of an extension type
-    # interned_cname   string     C name of interned name string
-    # pystring_cname   string     C name of Python version of string literal
-    # is_interned      boolean    For string const entries, value is interned
+    # name              string          Python name of entity
+    # cname             string          C name of entity
+    # type              PyrexType       Type of entity
+    # ctype             PyrexType       Declared C type, if different from Pyrex type
+    # doc               string          Doc string
+    # init              string          Initial value
+    # visibility        'private' or 'public' or 'extern'
+    # is_builtin        boolean         Is an entry in the Python builtins dict
+    # is_cglobal        boolean         Is a C global variable
+    # is_pyglobal       boolean         Is a Python module-level variable or
+    #                                     class attribute during class construction
+    # is_variable       boolean         Is a variable
+    # is_cfunction      boolean         Is a C function
+    # is_cmethod        boolean         Is a C method of an extension type
+    # is_builtin_method boolean         Is a method corresponding to a Python/C API func
+    # is_type           boolean         Is a type definition
+    # is_const          boolean         Is a constant
+    # is_property       boolean         Is a property of an extension type:
+    # doc_cname         string or None  C const holding the docstring
+    # getter_cname      string          C func for getting property
+    # setter_cname      string          C func for setting or deleting property
+    # is_self_arg       boolean         Is the "self" arg of an exttype method
+    # is_readonly       boolean         Can't be assigned to
+    # func_cname        string          C func implementing Python func
+    # pos               position        Source position where declared
+    # namespace_cname   string          If is_pyglobal, the C variable
+    #                                     holding its home namespace
+    # pymethdef_cname  string           PyMethodDef structure
+    # signature        Signature        Arg & return types for Python func
+    # init_to_none     boolean          True if initial value should be None
+    # as_variable      Entry            Alternative interpretation of extension
+    #                                     type name or builtin C function as a variable
+    # xdecref_cleanup  boolean          Use Py_XDECREF for error cleanup
+    # in_cinclude      boolean          Suppress C declaration code
+    # enum_values      [Entry]          For enum types, list of values
+    # qualified_name   string           "modname.funcname" or "modname.classname"
+    #                                     or "modname.classname.funcname"
+    # is_declared_generic  boolean        Is declared as PyObject * even though its
+    #                                       type is an extension type
+    # as_module        None             Module scope, if a cimported module
+    # is_inherited     boolean          Is an inherited attribute of an extension type
+    # #interned_cname   string          C name of interned name string
+    # pystring_cname   string           C name of Python version of string literal
+    # is_interned      boolean          For string const entries, value is interned
     # used             boolean
+    # is_special       boolean          Is a special method or property accessor
+    #                                     of an extension type
+    # defined_in_pxd   boolean          Is defined in a .pxd file (not just declared)
+    # api              boolean          Generate C API for C class or function
+    # utility_code     string           Utility code needed when this entry is used
 
     borrowed = 0
     init = ""
     visibility = 'private'
+    ctype = None
     is_builtin = 0
     is_cglobal = 0
     is_pyglobal = 0
     is_variable = 0
     is_cfunction = 0
     is_cmethod = 0
+    is_builtin_method = 0
     is_type = 0
     is_const = 0
     is_property = 0
@@ -89,11 +99,15 @@ class Entry:
     in_cinclude = 0
     as_module = None
     is_inherited = 0
-    interned_cname = None
+    #interned_cname = None
     pystring_cname = None
     is_interned = 0
     used = 0
-    
+    is_special = 0
+    defined_in_pxd = 0
+    api = 0
+    utility_code = None
+
     def __init__(self, name, cname, type, pos = None, init = None):
         self.name = name
         self.cname = cname
@@ -107,6 +121,7 @@ class Scope:
     # outer_scope       Scope or None      Enclosing scope
     # entries           {string : Entry}   Python name to entry, non-types
     # const_entries     [Entry]            Constant entries
+    # type_entries      [Entry]            Struct/union/enum/typedef/exttype entries
     # sue_entries       [Entry]            Struct/union/enum entries
     # arg_entries       [Entry]            Function argument entries
     # var_entries       [Entry]            User-defined variable entries
@@ -148,6 +163,7 @@ class Scope:
             self.scope_prefix = mangled_name
         self.entries = {}
         self.const_entries = []
+        self.type_entries = []
         self.sue_entries = []
         self.arg_entries = []
         self.var_entries = []
@@ -221,27 +237,42 @@ class Scope:
         return entry
     
     def declare_type(self, name, type, pos, 
-            cname = None, visibility = 'private'):
+            cname = None, visibility = 'private', defining = 1):
         # Add an entry for a type definition.
         if not cname:
             cname = name
         entry = self.declare(name, cname, type, pos)
         entry.visibility = visibility
         entry.is_type = 1
+        if defining:
+            self.type_entries.append(entry)
+        return entry
+    
+    def declare_typedef(self, name, base_type, pos, cname = None,
+            visibility = 'private'):
+        if not cname:
+            if self.in_cinclude or visibility == 'public':
+                cname = name
+            else:
+                cname = self.mangle(Naming.type_prefix, name)
+        type = PyrexTypes.CTypedefType(cname, base_type)
+        entry = self.declare_type(name, type, pos, cname, visibility)
+        type.qualified_name = entry.qualified_name
         return entry
         
     def declare_struct_or_union(self, name, kind, scope, 
-            typedef_flag, pos, cname = None):
+            typedef_flag, pos, cname = None, visibility = 'private'):
         # Add an entry for a struct or union definition.
         if not cname:
-            if self.in_cinclude:
+            if self.in_cinclude or visibility == 'public':
                 cname = name
             else:
                 cname = self.mangle(Naming.type_prefix, name)
         entry = self.lookup_here(name)
         if not entry:
             type = CStructOrUnionType(name, kind, scope, typedef_flag, cname)
-            entry = self.declare_type(name, type, pos, cname)
+            entry = self.declare_type(name, type, pos, cname,
+                visibility = visibility, defining = scope is not None)
             self.sue_entries.append(entry)
         else:
             if not (entry.is_type and entry.type.is_struct_or_union):
@@ -250,8 +281,11 @@ class Scope:
                 error(pos, "'%s' already defined" % name)
             else:
                 self.check_previous_typedef_flag(entry, typedef_flag, pos)
+                self.check_previous_visibility(entry, visibility, pos)
                 if scope:
+                    entry.pos = pos
                     entry.type.scope = scope
+                    self.type_entries.append(entry)
         if not scope and not entry.type.scope:
             self.check_for_illegal_incomplete_ctypedef(typedef_flag, pos)
         return entry
@@ -261,17 +295,24 @@ class Scope:
             error(pos, "'%s' previously declared using '%s'" % (
                 entry.name, ("cdef", "ctypedef")[entry.type.typedef_flag]))
     
-    def declare_enum(self, name, pos, cname, typedef_flag):
+    def check_previous_visibility(self, entry, visibility, pos):
+        if entry.visibility <> visibility:
+            error(pos, "'%s' previously declared as '%s'" % (
+                entry.name, entry.visibility))
+    
+    def declare_enum(self, name, pos, cname, typedef_flag,
+            visibility = 'private'):
         if name:
             if not cname:
-                if self.in_cinclude:
+                if self.in_cinclude or visibility == 'public':
                     cname = name
                 else:
                     cname = self.mangle(Naming.type_prefix, name)
             type = CEnumType(name, cname, typedef_flag)
         else:
-            type = c_int_type
-        entry = self.declare_type(name, type, pos, cname = cname)
+            type = PyrexTypes.c_anon_enum_type
+        entry = self.declare_type(name, type, pos, cname = cname,
+            visibility = visibility)
         entry.enum_values = []
         self.sue_entries.append(entry)
         return entry	
@@ -303,18 +344,32 @@ class Scope:
         self.pyfunc_entries.append(entry)
     
     def declare_cfunction(self, name, type, pos, 
-            cname = None, visibility = 'private', defining = 0):
+            cname = None, visibility = 'private', defining = 0, api = 0, in_pxd = 0):
         # Add an entry for a C function.
-        if not cname:
-            if visibility <> 'private':
-                cname = name
-            else:
-                cname = self.mangle(Naming.func_prefix, name)
-        entry = self.add_cfunction(name, type, pos, cname, visibility)
-        entry.func_cname = cname
+        entry = self.lookup_here(name)
+        if entry:
+            if visibility <> 'private' and visibility <> entry.visibility:
+                error(pos, "Function '%s' previously declared as '%s'" % (
+                    name, entry.visibility))
+            if not entry.type.same_as(type):
+                error(pos, "Function signature does not match previous declaration")
+        else:
+            if not cname:
+                if api or visibility <> 'private':
+                    cname = name
+                else:
+                    cname = self.mangle(Naming.func_prefix, name)
+            entry = self.add_cfunction(name, type, pos, cname, visibility)
+            entry.func_cname = cname
+        if in_pxd and visibility <> 'extern':
+            entry.defined_in_pxd = 1
+        if api:
+            entry.api = 1
+        if not defining and not in_pxd and visibility <> 'extern':
+            error(pos, "Non-extern C function declared but not defined")
         return entry
     
-    def add_cfunction(self, name, type, pos, cname, visibility):
+    def add_cfunction(self, name, type, pos, cname, visibility = 'private'):
         # Add a C function entry without giving it a func_cname.
         entry = self.declare(name, cname, type, pos)
         entry.is_cfunction = 1
@@ -322,6 +377,23 @@ class Scope:
         self.cfunc_entries.append(entry)
         return entry
     
+    def attach_var_entry_to_c_class(self, entry):
+        # The name of an extension class has to serve as both a type name and a
+        # variable name holding the type object. It is represented in the symbol
+        # table by a type entry with a variable entry attached to it. For the
+        # variable entry, we use a read-only C global variable whose name is an
+        # expression that refers to the type object.
+        var_entry = Entry(name = entry.name,
+            #type = py_object_type,
+            type = py_type_type,
+            pos = entry.pos,
+            #cname = "((PyObject*)%s)" % entry.type.typeptr_cname
+            cname = entry.type.typeptr_cname)
+        var_entry.is_variable = 1
+        var_entry.is_cglobal = 1
+        var_entry.is_readonly = 1
+        entry.as_variable = var_entry
+        
     def find(self, name, pos):
         # Look up name, report error if not found.
         entry = self.lookup(name)
@@ -448,15 +520,60 @@ class Scope:
 
 class BuiltinScope(Scope):
     #  The builtin namespace.
+    #
+    # type_names        {string : 1}       Set of type names (used during parsing)
     
     def __init__(self):
         Scope.__init__(self, "__builtin__", None, None)
+        self.type_names = {}
     
     def declare_builtin(self, name, pos):
         entry = self.declare(name, name, py_object_type, pos)
         entry.is_builtin = 1
         return entry
     
+    def declare_builtin_constant(self, name, type, cname, ctype = None):
+        entry = self.declare(name, cname, type, None)
+        if ctype:
+            entry.ctype = ctype
+        entry.is_variable = 1
+        entry.is_cglobal = 1
+        entry.is_readonly = 1
+        return entry
+    
+    def declare_builtin_cfunction(self, name, type, cname, python_equiv = None,
+            utility_code = None):
+        # If python_equiv == "*", the Python equivalent has the same name
+        # as the entry, otherwise it has the name specified by python_equiv.
+        entry = self.declare_cfunction(name, type, None, cname)
+        entry.utility_code = utility_code
+        if python_equiv:
+            if python_equiv == "*":
+                python_equiv = name
+            var_entry = Entry(python_equiv, python_equiv, py_object_type)
+            var_entry.is_variable = 1
+            var_entry.is_builtin = 1
+            entry.as_variable = var_entry
+        return entry
+    
+    def declare_builtin_class(self, name, objstruct_cname, typeobj_cname):
+        type = PyExtensionType(name, typedef_flag = 1, base_type = None)
+        type.module_name = "__builtin__"
+        type.typeptr_cname = "(&%s)" % typeobj_cname
+        type.objstruct_cname = objstruct_cname
+        scope = CClassScope(name = name, outer_scope = self, visibility = "extern")
+        type.set_scope(scope)
+        entry = self.declare_type(name, type, pos = None, visibility = "extern",
+            defining = 0)
+        self.attach_var_entry_to_c_class(entry)
+        self.type_names[name] = 1
+        return entry
+    
+    def find_type(self, name):
+        #  Used internally during initialisation, always succeeds
+        entry = self.lookup_here(name)
+        return entry.type
+
 
 class ModuleScope(Scope):
     # module_name          string             Python name of the module
@@ -474,13 +591,13 @@ class ModuleScope(Scope):
     # context              Context
     # parent_module        Scope              Parent in the import namespace
     # module_entries       {string : Entry}   For cimport statements
-    # type_names           {string : 1}       Set of type names (used during parsing)
     # pxd_file_loaded      boolean            Corresponding .pxd file has been processed
     # cimported_modules    [ModuleScope]      Modules imported with cimport
     # intern_map           {string : string}  Mapping from Python names to interned strs
     # interned_names       [string]           Interned names pending generation of declarations
     # all_pystring_entries [Entry]            Python string consts from all scopes
     # types_imported       {PyrexType : 1}    Set of types for which import code generated
+    # type_names           {string : 1}       Set of type names (used during parsing)
 
     def __init__(self, name, parent_module, context):
         self.parent_module = parent_module
@@ -499,7 +616,7 @@ class ModuleScope(Scope):
         self.module_entries = {}
         self.python_include_files = ["Python.h", "structmember.h"]
         self.include_files = []
-        self.type_names = {}
+        self.type_names = self.outer_scope.type_names.copy()
         self.pxd_file_loaded = 0
         self.cimported_modules = []
         self.intern_map = {}
@@ -515,7 +632,7 @@ class ModuleScope(Scope):
     
     def declare_builtin(self, name, pos):
         entry = Scope.declare_builtin(self, name, pos)
-        entry.interned_cname = self.intern(name)
+        #entry.interned_cname = self.intern(name)
         return entry
     
     def intern(self, name):
@@ -598,8 +715,8 @@ class ModuleScope(Scope):
                     "Non-cdef global variable is not a generic Python object")
             entry.is_pyglobal = 1
             entry.namespace_cname = self.module_cname
-            if Options.intern_names:
-                entry.interned_cname = self.intern(name)
+            #if Options.intern_names:
+            #	entry.interned_cname = self.intern(name)
         else:
             entry.is_cglobal = 1
             self.var_entries.append(entry)
@@ -634,10 +751,7 @@ class ModuleScope(Scope):
     
     def declare_c_class(self, name, pos, defining, implementing,
         module_name, base_type, objstruct_cname, typeobj_cname,
-        visibility, typedef_flag):
-        #
-        #print "declare_c_class:", name
-        #print "...visibility =", visibility
+        visibility, typedef_flag, api):
         #
         # Look for previous declaration as a type
         #
@@ -660,7 +774,8 @@ class ModuleScope(Scope):
             else:
                 type.module_name = self.qualified_name
             type.typeptr_cname = self.mangle(Naming.typeptr_prefix, name)
-            entry = self.declare_type(name, type, pos, visibility = visibility)
+            entry = self.declare_type(name, type, pos, visibility = visibility,
+                defining = 0)
             if objstruct_cname:
                 type.objstruct_cname = objstruct_cname
             elif not entry.in_cinclude:
@@ -680,6 +795,7 @@ class ModuleScope(Scope):
                 if base_type:
                     scope.declare_inherited_c_attributes(base_type.scope)
                 type.set_scope(scope)
+                self.type_entries.append(entry)
             else:
                 self.check_for_illegal_incomplete_ctypedef(typedef_flag, pos)
         else:
@@ -690,11 +806,15 @@ class ModuleScope(Scope):
         #
         # Fill in options, checking for compatibility with any previous declaration
         #
+        if defining:
+            entry.defined_in_pxd = 1
         if implementing:   # So that filenames in runtime exceptions refer to
             entry.pos = pos  # the .pyx file and not the .pxd file
-        if entry.visibility <> visibility:
-            error(pos, "Declaration of '%s' as '%s' conflicts with previous "
-                "declaration as '%s'" % (class_name, visibility, entry.visibility))
+        if visibility <> 'private' and entry.visibility <> visibility:
+            error(pos, "Class '%s' previously declared as '%s'"
+                % (name, entry.visibility))
+        if api:
+            entry.api = 1
         if objstruct_cname:
             if type.objstruct_cname and type.objstruct_cname <> objstruct_cname:
                 error(pos, "Object struct name differs from previous declaration")
@@ -772,22 +892,6 @@ class ModuleScope(Scope):
                 #print "ModuleScope.check_c_classes: allocating vtable cname for", self ###
                 type.vtable_cname = self.mangle(Naming.vtable_prefix, entry.name)
     
-    def attach_var_entry_to_c_class(self, entry):
-        # The name of an extension class has to serve as both a type
-        # name and a variable name holding the type object. It is
-        # represented in the symbol table by a type entry with a
-        # variable entry attached to it. For the variable entry,
-        # we use a read-only C global variable whose name is an
-        # expression that refers to the type object.
-        var_entry = Entry(name = entry.name,
-            type = py_object_type,
-            pos = entry.pos,
-            cname = "((PyObject*)%s)" % entry.type.typeptr_cname)
-        var_entry.is_variable = 1
-        var_entry.is_cglobal = 1
-        var_entry.is_readonly = 1
-        entry.as_variable = var_entry
-        
 
 class LocalScope(Scope):
 
@@ -849,6 +953,12 @@ class StructOrUnionScope(Scope):
             error(pos,
                 "C struct/union member cannot be declared %s" % visibility)
         return entry
+    
+    def declare_cfunction(self, name, type, pos, *args, **kwds):
+        error(pos, "C struct/union member cannot be a function")
+        #  Define it anyway to suppress further errors
+        kwds['defining'] = 1
+        Scope.declare_cfunction(self, name, type, pos, *args, **kwds)
 
 
 class ClassScope(Scope):
@@ -884,8 +994,8 @@ class PyClassScope(ClassScope):
             cname, visibility, is_cdef)
         entry.is_pyglobal = 1
         entry.namespace_cname = self.class_obj_cname
-        if Options.intern_names:
-            entry.interned_cname = self.intern(name)
+        #if Options.intern_names:
+        #	entry.interned_cname = self.intern(name)
         return entry
 
     def allocate_temp(self, type):
@@ -911,6 +1021,7 @@ class CClassScope(ClassScope):
     #  member_table_cname    string
     #  getset_table_cname    string
     #  has_pyobject_attrs    boolean  Any PyObject attributes?
+    #  pyattr_entries        [Entry]
     #  public_attr_entries   boolean  public/readonly attrs
     #  property_entries      [Entry]
     #  defined               boolean  Defined in .pxd file
@@ -926,6 +1037,7 @@ class CClassScope(ClassScope):
             self.member_table_cname = outer_scope.mangle(Naming.memtab_prefix, name)
             self.getset_table_cname = outer_scope.mangle(Naming.gstab_prefix, name)
         self.has_pyobject_attrs = 0
+        self.pyattr_entries = []
         self.public_attr_entries = []
         self.property_entries = []
         self.inherited_var_entries = []
@@ -938,6 +1050,11 @@ class CClassScope(ClassScope):
         return self.has_pyobject_attrs or \
             (self.parent_type.base_type and \
                 self.parent_type.base_type.scope.needs_gc())
+    
+    def declare_builtin_var(self, name, type, cname):
+        entry = self.declare(name, cname or name, type, None)
+        entry.is_variable = 1
+        return entry
 
     def declare_var(self, name, type, pos, 
             cname = None, visibility = 'private', is_cdef = 0):
@@ -956,8 +1073,9 @@ class CClassScope(ClassScope):
         entry.visibility = visibility
         entry.is_variable = 1
         self.var_entries.append(entry)
-        if type.is_pyobject:
+        if type.is_pyobject and name <> "__weakref__":
             self.has_pyobject_attrs = 1
+            self.pyattr_entries.append(entry)
         if visibility not in ('private', 'public', 'readonly'):
             error(pos,
                 "Attribute of extension type cannot be declared %s" % visibility)
@@ -976,18 +1094,33 @@ class CClassScope(ClassScope):
 
     def declare_pyfunction(self, name, pos):
         # Add an entry for a method.
+        if name == "__new__":
+            warning(pos, "__new__ method of extension type will change semantics "
+                "in a future version of Pyrex. Use __cinit__ instead.")
+            name = "__cinit__"
         entry = self.declare(name, name, py_object_type, pos)
         special_sig = get_special_method_signature(name)
         if special_sig:
+            entry.is_special = 1
             entry.signature = special_sig
             # Special methods don't get put in the method table
         else:
             entry.signature = pymethod_signature
             self.pyfunc_entries.append(entry)
         return entry
-            
+    
+    def lookup_here(self, name):
+        if name == "__new__":
+            name = "__cinit__"
+        return ClassScope.lookup_here(self, name)
+    
+    def declare_builtin_method(self, name, type, cname):
+        entry = ClassScope.add_cfunction(self, name, type, None, cname)
+        entry.is_builtin_method = 1
+        return entry
+    
     def declare_cfunction(self, name, type, pos,
-            cname = None, visibility = 'private', defining = 0):
+            cname = None, visibility = 'private', defining = 0, api = 0, in_pxd = 0):
         if get_special_method_signature(name):
             error(pos, "Special methods must be declared with 'def', not 'cdef'")
         args = type.args
@@ -1002,6 +1135,7 @@ class CClassScope(ClassScope):
             else:
                 if defining and entry.func_cname:
                     error(pos, "'%s' already defined" % name)
+                #print "CClassScope.declare_cfunction: checking signature" ###
                 if not entry.type.same_as(type, as_cmethod = 1):
                     error(pos, "Signature does not match previous declaration")
         else:
@@ -1059,6 +1193,7 @@ class PropertyScope(Scope):
         signature = get_property_accessor_signature(name)
         if signature:
             entry = self.declare(name, name, py_object_type, pos)
+            entry.is_special = 1
             entry.signature = signature
             return entry
         else:
